/*--------------------------------------------------------------------------
     "libfunc.h" - library functions and classes
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Aug.   5, 2002
                                        edited: May    1, 2004
                                        edited: Mar.  18, 2011
                                        edited: May   22, 2021
                                   last edited: Sep.  17, 2021
--------------------------------------------------------------------------*/

#ifndef _LIBFUNC_H_
#define _LIBFUNC_H_

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <algorithm>
#include <vector>
#include <deque>
#include <list>
#include <set>
#include <map>
#include <cctype>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <ctime>
using namespace std;

#ifndef true
#define true (-1)
#endif
#ifndef false
#define false 0
#endif
#ifndef M_PI
#define	M_PI    3.14159265358979323846
#endif
#define NORM 0
#define ERRE (-1)
#define REVSTR "thermNASA rev. 2021.09.17"

// ----- based on CODATA 2018 physical constants ---------------------------

#define KBOLTZ   1.380649e-23      // Boltzmann constant [J K-1] (exact)
#define NAVOG    6.02214076e+23    // Avogadro constant [mol-1]  (exact)
#define RGAS     8.314462618       // molar gas const [J K-1 mol-1] = KBOLTZ*NAVOG
#define RBCMMOL  83.14462618       // R_bar_cm_mol [bar K-1 cm3 mol-1] = RGAS*10
#define RBCMMLC  1.380649e-22      // R_bar_cm [bar K-1 cm3 molecule-1] = KBOLTZ*10
#define RGAS_CAL 1.987204259       // molar gas const [cal K-1 mol-1]

// ===== general purpose functions and classes =============================

void showRevision();
string strToUpper(const string &s);
string strToLower(const string &s);
string format(int v);
string format(int v, int len);
string format(string s, int len);
string formatLeft(string s, int len);
string formatRight(string s, int len);
string formatE(double v, int len, int prc);
string formatF(double v, int len, int prc);
string formatFnoDot(double v, int len);
string formatFautoLen(double v, int prc);
string formatD(int v, int len);
string formatBool(int flag);
int valBoolStr(string s);
long valHexStr(string s);
double atofx(const string &s);
string strCurDate();
int nextCorCket(string &s, int start);
int numMatchStrList(string s, vector<string> *vs);
ostream &operator<<(ostream &os, vector<string> &vs);
int splitToken(const string &s, const string &spl, vector<string> &vs);
string strTermBy(const string &s, char ct);
string strRemoveSpc(const string &s);
int getOneLine(ifstream &inf, string &s);
int getLineRmCmm(ifstream &inf, string &s, const string &cms);
int getLineApp(ifstream &inf, string &s);
int getTokensFrom(ifstream &inf, int ntkn, vector<string> &vs);
int getTokenFrom(ifstream &inf, string &s);
int getCharsFile(string flname, string &s);
string getBaseName(string &s);
string getBasePartEndWith(string nam, string strEnd);
int getGCM(int x, int y);
int getLCM(int x, int y);
int copyFile(string flnin, string flnout);
int addToFile(string flnin1, string flnin2, string flnout);
int add3Files(string flnin1, string flnin2, string flnin3, string flnout);

/*------------------------------------------------------------------------
     List of temperatures
------------------------------------------------------------------------*/

class tempList : public vector<double> {
 protected:
  enum errcode { FAILopen, EMPfile, INVkey, INVstep, INVrange, INVrngMANY,
                 INVparN, INVtemp };
 public:                         // --- member functions ---
  tempList();
  void setDefault();
  void setThermDefault();
  void setPAC99Default();
  int readFrom(string flnam);
  int addFrom(string slin);
  int setBy(string slin);
  void sortElimDup();
  int addRange(double st, double et, double dt);
  int addRangeVS(vector<string> &vs);
  int setByRange(double st, double et, double dt);
  int setByRangeVS(vector<string> &vs);
  int addRecipRange(double numr, double stt, double lst, double stp);
  int addRecipRangeVS(vector<string> &vs);
  int setByRecipRange(double numr, double stt, double lst, double stp);
  int setByRecipRangeVS(vector<string> &vs);
  int addGauChebGrd(double Tmin, double Tmax, int nT);
  int addGauChebGrdVS(vector<string> &vs);
  int setByGauChebGrd(double Tmin, double Tmax, int nT);
  int setByGauChebGrdVS(vector<string> &vs);
  int addList(vector<double> &vd);
  int addListVS(vector<string> &vs);
  int setByList(vector<double> &vd);
  int setByListVS(vector<string> &vs);
  ostream &print(ostream &os);
  double maxTemp();
 protected:
  int errms(errcode erc, string sv = "");
};                                        // --- globals ---
ostream &operator<<(ostream &os, tempList &tl);

#endif
