/*--------------------------------------------------------------------------
     "thermEquil.cc" - Calculation of equilibrium constants
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Oct.  19, 2005
                                        edited: Mar.  26, 2007
                                        edited: Mar.  14, 2010
                                        edited: Mar.  31, 2014
                                   last edited: May   22, 2021
----------------------------------------------------------------------------
 usage: thermEquil control.ext
--------------------------------------------------------------------------*/

#include "thermNASA.h"
using namespace std;

/* ===== main ========================================================= */

int main(int argc, char *argv[]) {
  int ntkns, flagOld = true, retc, itkn, it, tempRead = false;
  int nRct, nPrd, irct, iprd, nDelta, calc_reverse = false, conc_molecule;
  int energy_kelvin, nxp;
  double T, HRT, SR, CpR, deltaHRT, deltaSR, deltaGRT, Kp, Kcmol, Kcmlc;
  double invT, rateA, raten, rateEa, rateEaR, kfor, krev;
  string cinfln, basfln, outfln, slin, key, thmfln = "";
  string spu = "bar", scu_mol = "(mol.cm-3)", scu_mlc = "(molecule.cm-3)";
  string sxp, sKpu, sKcu_mol, sKcu_mlc, sciu, skfu, skru;
  vector<string> tkns, spRct, spPrd;
  ifstream inf;
  ofstream ofs;
  tempList temperat;
  thermTable thmCEA;
  thermTable::iterator p;
  vector<thermTable::iterator> prct, pprd;

  showRevision();
  if (argc != 2) { cout << "usage: thermEquil control.ext\n"; return 1; }
  cinfln = argv[1];
  basfln = getBaseName(cinfln);
  outfln = basfln + ".csv";

  inf.open(cinfln.c_str());
  if (!inf)
    { cout << "failed to open [" << cinfln << "].\n"; return 1; }
  while (getLineApp(inf, slin) == NORM) {
    ntkns = splitToken(slin, " ", tkns); key = tkns[0];
    if (key == "thermodata") {
      if (ntkns < 2 || ntkns > 3)
        { cout << "Invalid number of arguments for thermodata.\n"; return 1; }
      thmfln = tkns[1];
      if (ntkns == 3) {
        if (tkns[2] == "new") { flagOld = false; }
        else if (tkns[2] == "old") { flagOld = true; }
        else {
          cout << "Invalid argument [" << tkns[2] << "] for thermodata.\n";
          return 1;
        }
      }
    } else if (key.substr(0, 4) == "temp") {
      if (!tempRead) {
        if (temperat.setBy(slin) == ERRE) { return ERRE; }
        tempRead = true;
      } else {
        if (temperat.addFrom(slin) == ERRE) { return ERRE; }
      }
    } else if (key == "reactants") {
      if (ntkns < 2) { cout << "no argument for reactants.\n"; return 1; }
      spRct.clear();
      for (itkn = 1; itkn < ntkns; itkn++) { spRct.push_back(tkns[itkn]); }
      nRct = spRct.size();
    } else if (key == "products") {
      if (ntkns < 2) { cout << "no argument for products.\n"; return 1; }
      spPrd.clear();
      for (itkn = 1; itkn < ntkns; itkn++) { spPrd.push_back(tkns[itkn]); }
      nPrd = spPrd.size();
    } else if (key == "forward_k") {
      if (ntkns != 6)
        { cout << "Invalid number of arguments for forward_k.\n"; return 1; }
      if (tkns[1] == "mol/cm3") { conc_molecule = false; }
      else if (tkns[1] == "molecule/cm3") { conc_molecule = true; }
      else { cout << "Invalid argument " << tkns[1] << " for forward_k"; return 1; }
      if (tkns[2] == "cal/mol") { energy_kelvin = false; }
      else if (tkns[2] == "kelvin") { energy_kelvin = true; }
      else { cout << "Invalid argument " << tkns[2] << " for forward_k"; return 1; }
      rateA = atofx(tkns[3]); raten = atofx(tkns[4]); rateEa = atofx(tkns[5]);
      if (energy_kelvin) { rateEaR = rateEa; }
      else { rateEaR = rateEa / RGAS_CAL; }
      calc_reverse = true;
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return 1;
    }
  }
  inf.close();
  if (thmfln == "") { cout << "No thermodata.\n"; return 1; }
  if (nRct < 1) { cout << "No reactants.\n"; return 1; }
  if (nPrd < 1) { cout << "No products.\n"; return 1; }
  nDelta = nPrd - nRct;

  if (flagOld) { retc = thmCEA.appendOldFrom(thmfln); }
  else { retc = thmCEA.appendNewFrom(thmfln); }
  if (retc == ERRE)
    { cout << "thermodata file read error!\n"; return 1; }
  prct.clear(); pprd.clear();
  for (irct = 0; irct < nRct; irct++) {
    if ((p = thmCEA.find(spRct[irct])) == thmCEA.end())
      { cout << "unknown reactant!\n"; return 1; }
    prct.push_back(p);
  }
  for (iprd = 0; iprd < nPrd; iprd++) {
    if ((p = thmCEA.find(spPrd[iprd])) == thmCEA.end())
      { cout << "unknown product!\n"; return 1; }
    pprd.push_back(p);
  }

  ofs.open(outfln.c_str());
  ofs << "[ ";
  for (irct = 0; irct < nRct; irct++)
    { ofs << spRct[irct]; if (irct < nRct - 1) { ofs << '+'; } }
  ofs << " = ";
  for (iprd = 0; iprd < nPrd; iprd++)
    { ofs << spPrd[iprd]; if (iprd < nPrd - 1) { ofs << '+'; } }
  if (nDelta == 0) { sKpu = "-"; sKcu_mol = "-"; sKcu_mlc = "-"; }
  else if (nDelta == 1) { sKpu = spu; sKcu_mol = scu_mol; sKcu_mlc = scu_mlc; }
  else {
    sxp = format(nDelta); sKpu = spu + sxp;
    sKcu_mol = scu_mol + sxp; sKcu_mlc = scu_mlc + sxp;
  }
  ofs << " ]\nT / K,1000/T,delta-H/RT,delta-S/R,delta-G/RT,Kp / " << sKpu
      << ",Kc / " << sKcu_mol << ",Kc / " << sKcu_mlc;
  if (calc_reverse) {
    if (conc_molecule) { sciu = "(cm3.molecule-1)"; }
    else { sciu = "(cm3.mol-1)"; }
    if (nRct == 1) { skfu = "s-1"; }
    else if (nRct == 2) { skfu = sciu + "s-1"; }
    else { nxp = nRct - 1; sxp = format(nxp); skfu = sciu + sxp + "s-1"; }
    if (nPrd == 1) { skru = "s-1"; }
    else if (nPrd == 2) { skru = sciu + "s-1"; }
    else { nxp = nPrd - 1; sxp = format(nxp); skru = sciu + sxp + "s-1"; }
    ofs << ",kf / " << skfu << ",kr / " << skru;
  }
  ofs << endl;

  for (it = 0; it < temperat.size(); it++) {
    T = temperat[it]; invT = 1000. / T;
    deltaHRT = 0.; deltaSR = 0.;
    for (iprd = 0; iprd < nPrd; iprd++) {
      if (pprd[iprd]->second.getThermFuncs(T, HRT, SR, CpR) == ERRE)
        { cout << "thermodata calculation error!\n"; return 1; }
      deltaHRT += HRT; deltaSR += SR;
    }
    for (irct = 0; irct < nRct; irct++) {
      if (prct[irct]->second.getThermFuncs(T, HRT, SR, CpR) == ERRE)
        { cout << "thermodata calculation error!\n"; return 1; }
      deltaHRT -= HRT; deltaSR -= SR;
    }
    deltaGRT = deltaHRT - deltaSR;
    Kp = exp(-deltaGRT);
    Kcmol = Kp / pow(RBCMMOL * T, nDelta);
    Kcmlc = Kp / pow(RBCMMLC * T, nDelta);
    ofs << T << ',' << invT << ',' << deltaHRT << ',' << deltaSR << ','
      << deltaGRT << ',' << Kp << ',' << Kcmol << ',' << Kcmlc;
    if (calc_reverse) {
      kfor = rateA * pow(T, raten) * exp(-rateEaR / T);
      if (conc_molecule) { krev = kfor / Kcmlc; }
      else { krev = kfor / Kcmol; }
      ofs << ',' << kfor << ',' << krev;
    }
    ofs << endl;
  }
  ofs.close();
  return 0;
}

