/*--------------------------------------------------------------------------
     "thermLi97.cc" - Read thermodata and output to .i97 file
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Sept. 27, 2013
                                   last edited: May   21, 2021
----------------------------------------------------------------------------
 usage: thermLi97 input.ext [-new/-old]
--------------------------------------------------------------------------*/

#include "thermNASA.h"
using namespace std;

#define NTMP 29

double Tref = 298.15;
double Tlst[NTMP] = { 200., 298.15, 300., 400., 500., 600., 700., 800., 900.,
        1000., 1100., 1200., 1300., 1400., 1500., 1600., 1800., 2000., 2200.,
        2400., 2600., 2800., 3000., 3300., 3600., 3900., 4200., 4600., 5000. };

/* ===== main ========================================================= */

int main(int argc, char *argv[]) {
  int i, numatm, flagOld = true, retc;
  double HRTref, SRref, CpRref;
  double HRT, SR, CpR;
  string flninp, flnbas, flnout, sopt, namelm, chFormula;
  ofstream ofs;
  atmcCompSorted acs;
  atmcCompSorted::iterator a;
  thermTable thmCEA;
  thermTable::iterator p;

  showRevision();
  if ((argc < 2) || (argc > 3))
    { cout << "usage: thermLi97 input.ext [-new/-old]\n"; return 1; }
  flninp = argv[1]; flnbas = getBaseName(flninp);
  flnout = flnbas + ".i97";
  if (argc == 3) {
    sopt = argv[2];
    if (sopt == "-new") { flagOld = false; }
    else if (sopt == "-old") { flagOld = true; }
    else { cout << "usage: thermLi97 input.ext [-new/-old]\n"; return 1; }
  }
  if (flagOld) { retc = thmCEA.appendOldFrom(flninp); }
  else { retc = thmCEA.appendNewFrom(flninp); }
  if (retc == ERRE)
    { cout << "failed to read file [" << flninp << "].\n"; return 1; }

  ofs.open(flnout.c_str());
    if (!ofs) { cout << "failed to open [" << flnout << "].\n"; return 1; }
  for (p = thmCEA.begin(); p != thmCEA.end(); p++) {
    chFormula = "";
    acs.setFrom(p->second.atmCmp);
    for (a = acs.begin(); a != acs.end(); a++) {
      namelm = elmt.canonSymbol(a->second.first);
      if (namelm.length() > 1) { namelm[1] = toupper(namelm[1]); }
      numatm = a->second.second;
      chFormula += namelm + format(numatm);
    }
    chFormula += "                        ";
    if (p->second.getThermFuncs(Tref, HRTref, SRref, CpRref) == ERRE)
      { cout << "calculation error.\n"; ofs.close(); return 1; }
    ofs << "NAME  " << p->second.nameSpc << endl;
    ofs << chFormula.substr(0, 24) << "HF298 "
        << formatF(HRTref * RGAS * Tref, 12, 1) << "JOULES\n";
    ofs << "DATE  thLi97\n";
    ofs << "LSTS  OLD\n";
    ofs << "OUTP  MFIG              LSQS              ATM\n";
    ofs << "METH  READIN            KJOULE            BAR\n";
    for (i = 0; i < NTMP; i++) {
      if (p->second.getThermFuncs(Tlst[i], HRT, SR, CpR) == ERRE) {
        cout << "calculation error at T = " << formatF(Tlst[i], 12, 3) << endl;
      } else {
        ofs << "      T     " << formatF(Tlst[i], 12, 3)
            << "CP/R  " << formatF(CpR, 12, 4)
            << "H-H2RT" << formatF((HRT * Tlst[i] - HRTref * Tref) / Tlst[i] / 1000., 12, 7)
            << "S/R   " << formatF(SR, 12, 4) << endl;
      }
    }
    ofs << "FINISH\n";
  }
  ofs.close();
  return 0;
}

