/*--------------------------------------------------------------------------
     "thermNASA.cc" - thermodynamic functions in NASA polynomial form
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: June  17, 2005
                                        edited: Mar.  14, 2010
                                   last edited: May   20, 2021
--------------------------------------------------------------------------*/

#include "thermNASA.h"

// ***** thermNASA and related classes ***********************************

// ===== coefNASAset class ===============================================

coefNASAset::coefNASAset() { clear(); }           // ------- coefNASAset

void coefNASAset::clear()                         // ------------- clear
  { nCpCoef = 0; lowT = 0.; highT = 0.; dH298 = 0.; }

                                                  // ---- restoreOldFrom
int coefNASAset::restoreOldFrom(vector<string> &vs, double deflT,
 double defhT) {
  // Inputs - vs[0]: 2E10.0 (lowT, highT), vs[1]: 7E15.0 (a1-a7)
  int ic;
  nCpCoef = 5;
  for (ic = 0; ic < nCpCoef; ic++) { nx[ic] = ic; }
  lowT = atofx(vs[0].substr(0, 10));
  if (fabs(lowT - 290.) < 15.) { lowT = 270.; }
  if (lowT < 1e-10) { lowT = deflT; }
  highT = atofx(vs[0].substr(10, 10));
  if (highT < 1e-10) { highT = defhT; }
  for (ic = 0; ic < nCpCoef; ic++)
    { a[ic] = getValOldCoef(vs[1].substr(ic * 15, 15)); }
  for (ic = 0; ic < 2; ic++)
    { b[ic] = getValOldCoef(vs[1].substr((nCpCoef + ic) * 15, 15)); }
  double Tst = 298.15;
  if ((Tst >= lowT) && (Tst <= highT)) {
    double HR0 = b[0], HRT, SR, CpR;
    calcThermFuncs(Tst, HRT, SR, CpR);
    double HR298 = HRT * Tst;
    dH298 = (HR298 - HR0) * RGAS;
  }
  return NORM;
}

                                                  // ---- restoreNewFrom
int coefNASAset::restoreNewFrom(vector<string> &vs, double deflT,
 double defhT) {
  // Inputs - vs[0] to vs[2]: card-3 to card-5
  int ic;
  lowT = atofx(vs[0].substr(0, 11));
  if (lowT < 1e-10) { lowT = deflT; }
  highT = atofx(vs[0].substr(11, 11));
  if (highT < 1e-10) { highT = defhT; }
  nCpCoef = atoi(vs[0].substr(22, 1).c_str());
  for (ic = 0; ic < nCpCoef; ic++)
    { nx[ic] = atofx(vs[0].substr(ic * 5 + 23, 5)); }
  dH298 = atofx(vs[0].substr(65, 15));
  for (ic = 0; ic < 5; ic++)
    { a[ic] = atofx(vs[1].substr(ic * 16, 16)); }
  for (ic = 5; ic < nCpCoef; ic++)
    { a[ic] = atofx(vs[2].substr((ic - 5) * 16, 16)); }
  for (ic = 0; ic < 2; ic++)
    { b[ic] = atofx(vs[2].substr((ic + 3) * 16, 16)); }
  return NORM;
}

int coefNASAset::isTinInterval(double T) {        // ----- isTinInterval
  if ((T >= lowT) && (T <= highT)) { return true; }
  return false;
}

                                                  // ---- calcThermFuncs
void coefNASAset::calcThermFuncs(double T, double &HRT, double &SR,
 double &CpR) {
  int it;
  HRT = 0.; SR = 0.; CpR = 0.;
  for (it = 0; it < nCpCoef; it++) {
    double c = a[it], n = nx[it];
    CpR += c * pow(T, n);
    double tmHRT = c * pow(T, n);
    if (fabs(n + 1.) < 1e-10) { tmHRT *= log(T);  }
    else { tmHRT /= (n + 1.); }
    HRT += tmHRT;
    double tmSR = c * pow(T, n);
    if (fabs(n) < 1e-10) { tmSR *= log(T); }
    else { tmSR /= n; }
    SR += tmSR;
  }
  HRT += b[0] / T;
  SR += b[1];
}

ostream &coefNASAset::print(ostream &os) {        // ------------- print
  // print in new format
  int prec, ic;
  ios::fmtflags flgs;
  prec = os.precision(); flgs = os.flags();
  os.width(11); os.precision(3); os.setf(ios::fixed | ios::right);
  os << lowT; os.width(11); os << highT; os.width(1);
  os << nCpCoef; os.precision(1);
  for (ic = 0; ic < nCpCoef; ic++) { os.width(5); os << nx[ic]; }
  for (ic = nCpCoef; ic < 8; ic++) { os << "     "; }
  os << "  "; os.precision(3); os.width(15); os << dH298 << endl;
  os.precision(8); os.unsetf(ios::fixed); os.setf(ios::scientific);
  for (ic = 0; ic < 5; ic++) {
    if (ic < nCpCoef) { os.width(16); os << a[ic]; }
    else { os.width(16); os << 0.; }
  }
  os << endl;
  for (ic = 5; ic < 8; ic++) {
    if (ic < nCpCoef) { os.width(16); os << a[ic]; }
    else if (ic == 7) { os << "                "; }
    else { os.width(16); os << 0.; }
  }
  for (ic = 0; ic < 2; ic++) { os.width(16); os << b[ic]; }
  os << endl;
  os.flags(flgs); os.precision(prec);
  return os;
}

double coefNASAset::getValOldCoef(const string &s)   // -- getValOldCoef
  { string ws = fixOldCoefStr(s); return atofx(ws); }

string coefNASAset::fixOldCoefStr(const string &s) { // -- fixOldCoefStr
  static const string sfil = " 0.00000000e+00";
  int ic;
  string rets = s;
  for (ic = 1; ic < 15; ic ++)
    { if (rets[ic] == ' ') { rets[ic] = sfil[ic]; } }
  return rets;
}

ostream &operator<<(ostream &os, coefNASAset &cf)   // ------ operator<<
  { return cf.print(os); }

// ===== thermNASA class =================================================

thermNASA::thermNASA() { clear(); }               // --------- thermNASA

thermNASA::thermNASA(const thermNASA &tm) { (*this) = tm; }

void thermNASA::clear() {                         // ------------- clear
  vector<coefNASAset>::clear(); isUsed = false;
  optID = ""; atmCmp.clear(); wgtMol = 0.; H298 = 0.;
}

                                                  // ---- restoreOldFrom
int thermNASA::restoreOldFrom(vector<string> &vs, vector<double> &thT) {
  int ia;
  clear();
  nameSpc = strTermBy(vs[0].substr(0, 18), ' ');
  if (nameSpc.length() < 1) { return ERRE; }
  comments = strRemoveSpc(vs[0].substr(18, 6));
  for (ia = 0; ia < 5; ia++) {
    int pos = ia * 5 + 24;
    if (ia == 4) { pos += 29; }
    string aSym = strToUpper(strRemoveSpc(vs[0].substr(pos, 2)));
    int atmcNbr = elmt.atmcNbrBySymb(aSym);
    int numCnt = atoi(strRemoveSpc(vs[0].substr(pos + 2, 3)).c_str());
    if ((atmcNbr > 0) && (numCnt > 0))
      { atmCmp.insert(make_pair(atmcNbr, numCnt)); }
  }
  char sphse = strToUpper(vs[0].substr(44, 1))[0];
  switch (sphse) {
    case 'L': phase = 1; break;
    case 'S': phase = 2; break;
    default: phase = 0; break;
  }
  string sTlow = vs[0].substr(45, 10), sThigh = vs[0].substr(55, 10),
         sTcom = "  " + vs[0].substr(65, 8);
  vector<string> ett;
  coefNASAset cof;
  ett.clear(); ett.push_back(sTlow + sTcom);
  ett.push_back(vs[2].substr(30, 45) + vs[3].substr(0, 60));
  cof.clear(); cof.restoreOldFrom(ett, thT[0], thT[1]);
  push_back(cof);
  ett.clear(); ett.push_back(sTcom + sThigh);
  ett.push_back(vs[1].substr(0, 75) + vs[2].substr(0, 30));
  cof.clear(); cof.restoreOldFrom(ett, thT[1], thT[2]);
  push_back(cof);
  double Tst = 298.15, HRT, SR, CpR;
  if (getThermFuncs(Tst, HRT, SR, CpR) == NORM)
    { H298 = HRT * RGAS * Tst; }
  return NORM;
}

                                                  // ---- restoreNewFrom
int thermNASA::restoreNewFrom(vector<string> &vs, vector<double> &thT) {
  int ia, it, jr;
  clear();
  nameSpc = strRemoveSpc(vs[0].substr(0, 15));
  if (nameSpc.length() < 1) { return ERRE; }
  comments = strRemoveSpc(vs[0].substr(15, 65));
  int numTint = atoi(vs[1].substr(0, 2).c_str());
  optID = strRemoveSpc(vs[1].substr(3, 6));
  for (ia = 0; ia < 5; ia++) {
    int pos = ia * 8 + 10;
    string aSym = strToUpper(strRemoveSpc(vs[1].substr(pos, 2)));
    int atmcNbr = elmt.atmcNbrBySymb(aSym);
    int numCnt = atofx(vs[1].substr(pos + 2, 6));
    if ((atmcNbr > 0) && (numCnt > 0))
      { atmCmp.insert(make_pair(atmcNbr, numCnt)); }
  }
  phase = atoi(vs[1].substr(50, 2).c_str());
  wgtMol = atofx(vs[1].substr(52, 13));
  H298 = atofx(vs[1].substr(65, 15));
  for (it = 0; it < numTint; it++) {
    int irec = it * 3 + 2;
    vector<string> ett;
    coefNASAset cof;
    double dlT = 0., dhT = 0.;
    ett.clear();
    for (jr = 0; jr < 3; jr++) { ett.push_back(vs[irec + jr]); }
    if (it < 3) { dlT = thT[it]; dhT = thT[it + 1]; }
    else if (it == 3) { dlT = thT[it]; }
    cof.clear(); cof.restoreNewFrom(ett, dlT, dhT);
    push_back(cof);
  }
  return NORM;
}

                                                  // ----- getThermFuncs
int thermNASA::getThermFuncs(double T, double &HRT, double &SR,
 double &CpR) {
  int found = false;
  iterator pit;
  for (pit = begin(); pit != end(); pit++)
    { if (pit->isTinInterval(T)) { found = true; break; } }
  if (!found) { return ERRE; }
  pit->calcThermFuncs(T, HRT, SR, CpR);
  return NORM;
}

ostream &thermNASA::print(ostream &os) {          // ------------- print
  // print in new format
  int prec, ia, it;
  ios::fmtflags flgs;
  atmcComp::iterator ap;
  string aSym;
  double nAtm;
  prec = os.precision(); flgs = os.flags();
  os.width(15); os.setf(ios::left); os << nameSpc;
  os.width(65); os << comments << endl;
  os.width(2); os.setf(ios::right); os << size() << ' ';
  os.unsetf(ios::right);
  os.width(6); os.setf(ios::left); os << optID << ' ';
  for (ap = atmCmp.begin(); ap != atmCmp.end(); ap++) {
    aSym = elmt[ap->first].atomicSymbol;
    nAtm = ap->second;
    os.width(2); os.unsetf(ios::right); os.setf(ios::left); os << aSym;
    os.width(6); os.setf(ios::right | ios::fixed); os.precision(2);
    os << nAtm;
  }
  for (ia = atmCmp.size(); ia < 5; ia++) { os << "        "; }
  os.width(2); os << phase;
  os.width(13); os.precision(7); os << wgtMol;
  os.width(15); os.precision(3); os << H298 << endl;
  os.flags(flgs); os.precision(prec);
  for (it = 0; it < size(); it++) { os << (*this)[it]; }
  return os;
}

ostream &operator<<(ostream &os, thermNASA &th)   // -------- operator<<
  { return th.print(os); }

// ===== thermTable class ================================================

thermTable::thermTable() { clear(); }             // -------- thermTable

void thermTable::clear() {
  map<string, thermNASA>::clear();
  comments = ""; thresT.clear();
  double Tth = 200.;
  thresT.push_back(Tth);
  Tth = 1000.; thresT.push_back(Tth);
  Tth = 6000.; thresT.push_back(Tth);
  Tth = 20000.; thresT.push_back(Tth);
}

void thermTable::setM() {                         // -------------- setM
  string nam;
  thermNASA thm;
  clear();
  nam = "m"; thm.nameSpc = nam; insert(make_pair(nam, thm));
  nam = "M"; thm.nameSpc = nam; insert(make_pair(nam, thm));
}

int thermTable::appendFrom(string flname)         // -------- appendFrom
  { return appendOldFrom(flname); }

int thermTable::appendOldFrom(string flname) {    // ----- appendOldFrom
  int nbefore = size();
  if (appendOld(flname) == ERRE) { return ERRE; }
  int nafter = size();
  cout << nafter - nbefore << " thermodata loaded from ["
       << flname << "]\n";
  return NORM;
}

int thermTable::appendNewFrom(string flname) {    // ----- appendNewFrom
  int nbefore = size();
  if (appendNew(flname) == ERRE) { return ERRE; }
  int nafter = size();
  cout << nafter - nbefore << " thermodata (new form) loaded from ["
       << flname << "]\n";
  return NORM;
}

void thermTable::setAllIsUsed() {                 // ------ setAllIsUsed
  iterator p;
  for (p = begin(); p != end(); p++) { p->second.isUsed = true; }
}

set<int> thermTable::getUsedAtoms() {             // ------ getUsedAtoms
  atmcComp ac;
  atmcComp::iterator pa;
  set<int> aUsd;
  ac.clear();
  for (iterator p = begin(); p != end(); p++)
    { if (p->second.isUsed) { ac += p->second.atmCmp; } }
  aUsd.clear();
  for (pa = ac.begin(); pa != ac.end(); pa++) { aUsd.insert(pa->first); }
  return aUsd;
}

ostream &thermTable::print(ostream &os) {         // ------------- print
  // print in new format
  int prec, it;
  ios::fmtflags flgs;
  prec = os.precision(); flgs = os.flags();
  os << "thermo\n";
  os.precision(2); os.setf(ios::fixed | ios::right);
  for (it = 0; it < thresT.size(); it++)
    { os.width(10); os << thresT[it]; }
  for (it = thresT.size(); it < 4; it++) { os << "          "; }
  os.unsetf(ios::right); os.setf(ios::left); os.width(10);
  os << comments << endl;
  os.flags(flgs); os.precision(prec);
  for (iterator p = begin(); p != end(); p++) { os << p->second; }
  os << "END PRODUCTS\n";
  return os;
}

int thermTable::appendOld(string flname) {        // --------- appendOld
  int it;
  static const string spc30 = "                              ";
  string s;
  ifstream ifs(flname.c_str());
  if (!ifs) { return ERRE; }
  // read-in until thermo block header
  int headerFound = false;
  do {
    if (getThermLine(ifs, s) == ERRE) { return ERRE; }
    if (s.length() >= 6) {
      if (strToUpper(s).substr(0, 6) == "THERMO")
        { headerFound = true; }
    }
  } while (!headerFound);
  // next non-comment line should be temperature interval
  if (getThermLine(ifs, s) == ERRE) { return ERRE; }
  int lens = s.length();
  if (lens < 30) { s += spc30.substr(0, 30 - lens); }
  for (it = 0; it < 3; it++)
    { thresT[it] = atofx(s.substr(it * 10, 10)); }
  // following 4 lines repeated until footer are thermodata
  int thermoEnd = false, il;
  vector<string> vs;
  string nam;
  thermNASA thm;
  do {
    // eof without footer is allowed
    if (getThermLine(ifs, s) == ERRE) { thermoEnd = true; break; }
    if (strToUpper(s).substr(0, 3) == "END") { thermoEnd = true; }
    else {
      vs.clear(); vs.push_back(s);
      for (il = 0; il < 3; il++) {
        if (getThermLine(ifs, s) == ERRE) { return ERRE; }
        vs.push_back(s);
      }
      thm.restoreOldFrom(vs, thresT);
      nam = thm.nameSpc;
      insert(make_pair(nam, thm));
    }
  } while (!thermoEnd);
  ifs.close();
  return NORM;
}

int thermTable::appendNew(string flname) {     // ------------ appendNew
  int it;
  static const string spc40 = "                                        ";
  string s;
  ifstream ifs(flname.c_str());
  if (!ifs) { return ERRE; }
  // read-in until thermo block header
  int headerFound = false;
  do {
    if (getThermLine(ifs, s) == ERRE) { return ERRE; }
    if (s.length() >= 6) {
      if (strToUpper(s).substr(0, 6) == "THERMO")
        { headerFound = true; }
    }
  } while (!headerFound);
  // next non-comment line should be temperature interval
  if (getThermLine(ifs, s) == ERRE) { return ERRE; }
  int lens = s.length();
  if (lens < 40) { s += spc40.substr(0, 40 - lens); }
  for (it = 0; it < 4; it++)
    { thresT[it] = atofx(s.substr(it * 10, 10)); }
  if (lens > 40) { comments = strRemoveSpc(s.substr(40)); }
  // following 2 + n * 3 lines repeated until footer are thermodata
  int thermoEnd = false, nTintv, il;
  vector<string> vs;
  string nam;
  thermNASA thm;
  do {
    // eof without footer is allowed
    if (getThermLine(ifs, s) == ERRE) { thermoEnd = true; break; }
    if (strToUpper(s).substr(0, 3) == "END") { thermoEnd = true; }
    else {
      vs.clear(); vs.push_back(s);
      if (getThermLine(ifs, s) == ERRE) { return ERRE; }
      vs.push_back(s);
      nTintv = atoi(s.substr(0, 2).c_str());
      for (it = 0; it < nTintv; it++) {
        for (il = 0; il < 3; il++) {
          if (getThermLine(ifs, s) == ERRE) { return ERRE; }
          vs.push_back(s);
        }
      }
      thm.restoreNewFrom(vs, thresT);
      nam = thm.nameSpc;
      insert(make_pair(nam, thm));
    }
  } while (!thermoEnd);
  ifs.close();
  return NORM;
}

                                                  // ------ getThermLine
int thermTable::getThermLine(ifstream &ifs, string &s) {
  string sn;
  while (getOneLine(ifs, s) == NORM) {
    sn = strRemoveSpc(s);
    if (sn.length() > 0) {
      if (sn[0] != '!') { return NORM; }
    }
  }
  return ERRE;
}

ostream &operator<<(ostream &os, thermTable &tt)  // -------- operator<<
  { return tt.print(os); }

