/*--------------------------------------------------------------------------
     "thermRead.cc" - Read out thermodynamic data
                        (c) 2002-2021 by A. Miyoshi, Hiroshima University
                                       created: Oct.  19, 2005
                                        edited: Mar.  26, 2007
                                        edited: Jan.   2, 2013
                                        edited: May    5, 2015
                                        edited: May   22, 2021
                                   last edited: Sep.  17, 2021
                                           - added Cp/R output
----------------------------------------------------------------------------
 usage: thermRead control.ext
--------------------------------------------------------------------------*/

#include "thermNASA.h"
using namespace std;

double Tref = 298.15;

/* ===== main ========================================================= */

int main(int argc, char *argv[]) {
  int ntkns, flagOld = true, retc, it, tempRead = false;
  double T, HRT, SR, CpR;
  string cinfln, basfln, outfln, slin, key, thmfln = "", namspc;
  vector<string> tkns;
  ifstream inf;
  ofstream ofs;
  tempList temperat;
  thermTable thmCEA;
  thermTable::iterator p;

  showRevision();
  if (argc != 2) { cout << "usage: thermRead control.ext\n"; return 1; }
  cinfln = argv[1];
  basfln = getBaseName(cinfln);
  outfln = basfln + ".csv";

  inf.open(cinfln.c_str());
  if (!inf)
    { cout << "failed to open [" << cinfln << "].\n"; return 1; }
  while (getLineApp(inf, slin) == NORM) {
    ntkns = splitToken(slin, " ", tkns); key = tkns[0];
    if (key == "thermodata") {
      if (ntkns < 2 || ntkns > 3)
        { cout << "Invalid number of arguments for thermodata.\n"; return 1; }
      thmfln = tkns[1];
      if (ntkns == 3) {
        if (tkns[2] == "new") { flagOld = false; }
        else if (tkns[2] == "old") { flagOld = true; }
        else {
          cout << "Invalid argument [" << tkns[2] << "] for thermodata.\n";
          return 1;
        }
      }
    } else if (key.substr(0, 4) == "temp") {
      if (!tempRead) {
        if (temperat.setBy(slin) == ERRE) { return ERRE; }
        tempRead = true;
      } else {
        if (temperat.addFrom(slin) == ERRE) { return ERRE; }
      }
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return 1;
    }
  }
  inf.close();
  if (thmfln == "") { cout << "No thermodata.\n"; return 1; }

  if (flagOld) { retc = thmCEA.appendOldFrom(thmfln); }
  else { retc = thmCEA.appendNewFrom(thmfln); }
  if (retc == ERRE)
    { cout << "thermodata file read error!\n"; return 1; }

  ofs.open(outfln.c_str());
    if (!ofs) { cout << "failed to open [" << outfln << "].\n"; return 1; }
  for (p = thmCEA.begin(); p != thmCEA.end(); p++) {
    namspc = p->second.nameSpc;
    ofs << "," << namspc << "," << namspc << "," << namspc << "," << namspc;
  }
  ofs << "\nT / K";
  for (p = thmCEA.begin(); p != thmCEA.end(); p++)
    { ofs << ",H/RT,S/R,Cp/R,G/RT"; }
  ofs << endl;

  for (it = 0; it < temperat.size(); it++) {
    T = temperat[it]; ofs << T;
    for (p = thmCEA.begin(); p != thmCEA.end(); p++) {
      if (p->second.getThermFuncs(T, HRT, SR, CpR) == ERRE) {
        ofs << ",err,err,err";
      } else {
        ofs << "," << HRT << "," << SR << "," << CpR << "," << (HRT - SR);
      }
    }
    ofs << endl;
  }

  ofs << "T / K";
  for (p = thmCEA.begin(); p != thmCEA.end(); p++)
    { ofs << ",H / kJ.mol-1,S / J.K-1.mol-1,Cp / J.K-1.mol-1,G / kJ.mol-1"; }
  ofs << endl;
  T = Tref; ofs << T;
  for (p = thmCEA.begin(); p != thmCEA.end(); p++) {
    if (p->second.getThermFuncs(T, HRT, SR, CpR) == ERRE) {
      ofs << ",err,err,err";
    } else {
      ofs << "," << (HRT * RGAS * T / 1000.) << "," << (SR * RGAS) << ","
        << (CpR * RGAS) << "," << ((HRT - SR) * RGAS * T / 1000.);
    }
  }
  ofs << endl;

  ofs.close();
  return 0;
}

